/**\file jdoom.h
 *\section License
 * License: GPL
 * Online License Link: http://www.gnu.org/licenses/gpl.html
 *
 *\author Copyright © 2006-2013 Jaakko Keränen <jaakko.keranen@iki.fi>
 *\author Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

/**
 * All jDoom Headers
 *
 * A convenience header for including all jDoom headers.
 */

#ifndef __JDOOM_CONVENIENCE_H__
#define __JDOOM_CONVENIENCE_H__

#include "../../doom/include/acfnlink.h"
#include "../../doom/include/d_api.h"
#include "../../doom/include/d_config.h"
#include "../../doom/include/d_console.h"
#include "../../doom/include/d_event.h"
#include "../../doom/include/d_main.h"
#include "../../doom/include/d_refresh.h"
#include "../../doom/include/doomdef.h"
#include "../../doom/include/doomtype.h"
#include "../../doom/include/dstrings.h"
#include "../../doom/include/g_game.h"
#include "../../doom/include/info.h"
#include "../../doom/include/intermission.h"
#include "../../doom/include/m_cheat.h"
#include "../../doom/include/m_random.h"
#include "../../doom/include/p_enemy.h"
#include "../../doom/include/bossbrain.h"
#include "../../doom/include/p_inter.h"
#include "../../doom/include/p_lights.h"
#include "../../doom/include/p_local.h"
#include "../../doom/include/p_maputl.h"
#include "../../doom/include/p_mobj.h"
#include "../../doom/include/p_pspr.h"
#include "../../doom/include/p_setup.h"
#include "../../doom/include/p_spec.h"
#include "../../doom/include/p_telept.h"
#include "../../doom/include/r_defs.h"
#include "../../doom/include/st_stuff.h"
#include "../../doom/include/tables.h"

#endif // __JDOOM_CONVENIENCE_H__
