/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiTextureOp {
  aiTextureOp_Multiply = 0x0,
  aiTextureOp_Add = 0x1,
  aiTextureOp_Subtract = 0x2,
  aiTextureOp_Divide = 0x3,
  aiTextureOp_SmoothAdd = 0x4,
  aiTextureOp_SignedAdd = 0x5
}
