unit aiMatrix3x3;

interface

type TaiMatrix3x3 = packed record
   a1, a2, a3, a4: single;
   b1, b2, b3, b4: single;
   c1, c2, c3, c4: single;
end;
PaiMatrix3x3 = ^TaiMatrix3x3;



implementation

end.
