/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiReturn {
  aiReturn_SUCCESS = 0x0,
  aiReturn_FAILURE = -0x1,
  aiReturn_OUTOFMEMORY = -0x3,
  _AI_ENFORCE_ENUM_SIZE = 0x7fffffff
}
