/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiTextureMapMode {
  aiTextureMapMode_Wrap = 0x0,
  aiTextureMapMode_Clamp = 0x1,
  aiTextureMapMode_Decal = 0x3,
  aiTextureMapMode_Mirror = 0x2
}
