

// #ifndef ASSIMP_BUILD_SINGLETHREADED
// #	include <boost/thread.hpp>
// #endif

// We need to be sure to have the same STL settings as Assimp
#include <AssimpPCH.h>
#include <assimp/cimport.h>

// CPPUNIT
#include <cppunit/TestFixture.h>
#include <cppunit/extensions/HelperMacros.h>

#include <math.h>

#undef min
#undef max