LIBRARY			LZSS
DESCRIPTION		"Lempel-Ziv Compressed File I/O"

EXPORTS
	lzPassword		@1
	lzOpen			@2
	lzClose			@3
	lzOpenChunk		@4
	lzCloseChunk	@5
	lzSeek			@6
	lzGetW			@7
	lzGetL			@8
	lzPutW			@9
	lzPutL			@10
	lzGetWm			@11
	lzGetLm			@12
	lzPutWm			@13
	lzPutLm			@14
	lzRead			@15
	lzWrite			@16
	lzGetS			@17
	lzPutS			@18
	lzGetC			@19
	lzPutC			@20
