/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiTextureType {
  aiTextureType_NONE = 0x0,
  aiTextureType_DIFFUSE = 0x1,
  aiTextureType_SPECULAR = 0x2,
  aiTextureType_AMBIENT = 0x3,
  aiTextureType_EMISSIVE = 0x4,
  aiTextureType_HEIGHT = 0x5,
  aiTextureType_NORMALS = 0x6,
  aiTextureType_SHININESS = 0x7,
  aiTextureType_OPACITY = 0x8,
  aiTextureType_DISPLACEMENT = 0x9,
  aiTextureType_LIGHTMAP = 0xA,
  aiTextureType_REFLECTION = 0xB,
  aiTextureType_UNKNOWN = 0xC
}
