LIBRARY			AUDIO_WINMM

; All exports should be prefixed DS_.

EXPORTS

deng_LibraryType
deng_API

DS_Init
DS_Shutdown
DS_Event

DM_Music_Init
DM_Music_Update
DM_Music_Set
DM_Music_Get
DM_Music_Pause
DM_Music_Stop
DM_Music_Play
DM_Music_SongBuffer

DM_CDAudio_Init
DM_CDAudio_Update
DM_CDAudio_Set
DM_CDAudio_Get
DM_CDAudio_Pause
DM_CDAudio_Stop
DM_CDAudio_Play
