/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiShadingMode {
  aiShadingMode_Flat = 0x1,
  aiShadingMode_Gouraud = 0x2,
  aiShadingMode_Phong = 0x3,
  aiShadingMode_Blinn = 0x4,
  aiShadingMode_Toon = 0x5,
  aiShadingMode_OrenNayar = 0x6,
  aiShadingMode_Minnaert = 0x7,
  aiShadingMode_CookTorrance = 0x8,
  aiShadingMode_NoShading = 0x9,
  aiShadingMode_Fresnel = 0xa
}
