/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiPrimitiveType {
  aiPrimitiveType_POINT = 0x1,
  aiPrimitiveType_LINE = 0x2,
  aiPrimitiveType_TRIANGLE = 0x4,
  aiPrimitiveType_POLYGON = 0x8
}
