/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiTextureMapping {
  aiTextureMapping_UV = 0x0,
  aiTextureMapping_SPHERE = 0x1,
  aiTextureMapping_CYLINDER = 0x2,
  aiTextureMapping_BOX = 0x3,
  aiTextureMapping_PLANE = 0x4,
  aiTextureMapping_OTHER = 0x5
}
