/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiTextureFlags {
  aiTextureFlags_Invert = 0x1,
  aiTextureFlags_UseAlpha = 0x2,
  aiTextureFlags_IgnoreAlpha = 0x4
}
