/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiComponent {
  aiComponent_NORMALS = 0x2,
  aiComponent_TANGENTS_AND_BITANGENTS = 0x4,
  aiComponent_COLORS = 0x8,
  aiComponent_TEXCOORDS = 0x10,
  aiComponent_BONEWEIGHTS = 0x20,
  aiComponent_ANIMATIONS = 0x40,
  aiComponent_TEXTURES = 0x80,
  aiComponent_LIGHTS = 0x100,
  aiComponent_CAMERAS = 0x200,
  aiComponent_MESHES = 0x400,
  aiComponent_MATERIALS = 0x800
}
