/*
 * Copyright (c) 2002-2011 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __AMETHYST_CONTEXT_RELATION_H__
#define __AMETHYST_CONTEXT_RELATION_H__

#include "outputcontext.h"

class ContextRelation : public Linkable
{
public:
    ContextRelation(OutputContext *source = 0, OutputContext *target = 0);
    ContextRelation(const ContextRelation &other);

    ContextRelation *next() { return (ContextRelation*) Linkable::next(); }
    ContextRelation *prev() { return (ContextRelation*) Linkable::prev(); }

    OutputContext *source() { return _start; }
    OutputContext *target() { return _end; }

protected:
    OutputContext *_start, *_end;
};

#endif
