LIBRARY			AUDIO_DIRECTSOUND

; All exports should be prefixed DS_.

EXPORTS

deng_LibraryType
deng_API

DS_Init
DS_Shutdown
DS_Event

DS_SFX_Init
DS_SFX_CreateBuffer
DS_SFX_DestroyBuffer
DS_SFX_Load
DS_SFX_Reset
DS_SFX_Play
DS_SFX_Stop
DS_SFX_Refresh
DS_SFX_Set
DS_SFX_Setv
DS_SFX_Listener
DS_SFX_Listenerv
