/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 2.0.8
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


public enum aiDefaultLogStream {
  aiDefaultLogStream_FILE = 0x1,
  aiDefaultLogStream_STDOUT = 0x2,
  aiDefaultLogStream_STDERR = 0x4,
  aiDefaultLogStream_DEBUGGER = 0x8,
  _AI_DLS_ENFORCE_ENUM_SIZE = 0x7fffffff
}
