LIBRARY			AUDIO_FMOD

; All exports should be prefixed DS_.

EXPORTS

deng_LibraryType
deng_API

DS_Init
DS_Shutdown
DS_Event
DS_Set

; Sfx interface
DS_SFX_Init
DS_SFX_CreateBuffer
DS_SFX_DestroyBuffer
DS_SFX_Load
DS_SFX_Reset
DS_SFX_Play
DS_SFX_Stop
DS_SFX_Refresh
DS_SFX_Set
DS_SFX_Setv
DS_SFX_Listener
DS_SFX_Listenerv
DS_SFX_Getv

; Music interface
DM_Music_Init
DM_Music_Update
DM_Music_Get
DM_Music_Set
DM_Music_Pause
DM_Music_Stop
DM_Music_SongBuffer
DM_Music_Play
DM_Music_PlayFile

; CD audio interface
DM_CDAudio_Init
DM_CDAudio_Update
DM_CDAudio_Set
DM_CDAudio_Get
DM_CDAudio_Pause
DM_CDAudio_Stop
DM_CDAudio_Play
